package org.verohallinto.apitamoclient;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.cxf.Bus;
import org.apache.cxf.configuration.jsse.TLSClientParameters;
import org.apache.cxf.endpoint.Client;
import org.apache.cxf.ext.logging.LoggingInInterceptor;
import org.apache.cxf.ext.logging.LoggingOutInterceptor;
import org.apache.cxf.frontend.ClientProxy;
import org.apache.cxf.interceptor.Fault;
import org.apache.cxf.message.Message;
import org.apache.cxf.transport.http.HTTPConduit;
import org.apache.cxf.transports.http.configuration.ConnectionType;
import org.apache.cxf.transports.http.configuration.HTTPClientPolicy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.verohallinto.apitamoclient.dto.ResponseDTO;
import org.verohallinto.apitamoclient.sec.TLSParameters;
import org.verohallinto.apitamoclient.util.Utils;
import org.verohallinto.ilmoitin.gen.cxf.apitamopki.*;
import org.verohallinto.ilmoitin.gen.jaxb.tamo.en.CheckupResult;
import org.verohallinto.ilmoitin.gen.jaxb.tamo.en.TamoResult;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.Unmarshaller;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.soap.SOAPFault;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.soap.SOAPBinding;
import javax.xml.ws.soap.SOAPFaultException;
import java.io.*;
import java.net.URL;
import java.util.*;
import java.util.regex.Pattern;

/**
 * <p>CXF SOAP client.</p>
 */
public class ApitamopkiClient {

    private static final Logger log = LoggerFactory.getLogger(ApitamopkiClient.class);
    private static final String EMAIL_PATTERN;
    //
    private ApitamopkiType port = null;
    // WSDL
    private URL wsdl;
    // KeyStore
    private String ksJks;
    private String ksPwd;
    private String ksAlias;
    private String ksAliasPwd;
    // TrustStore
    private String tsJks;
    private String tsPwd;
    // Accepted hostname
    private String acceptHost;
    // Timeouts
    private long connTimeout;
    private long readTimeout;

    static {
        EMAIL_PATTERN = "^[_A-Za-z0-9-]+(\\.[_A-Za-z0-9-]+)*@[A-Za-z0-9-]+(\\.[A-Za-z0-9-]+)*((\\.[A-Za-z]{2,}){1}$)";
    }

    /**
     * Constructor.
     * </p>
     *
     * @param wsdl        <code>URL</code> service WSDL.
     * @param ksJks       <code>String</code> KeyStore JKS.
     * @param ksPwd       <code>String</code> KeyStore password.
     * @param ksAlias     <code>String</code> KeyStore alias name.
     * @param ksAliasPwd  <code>String</code> KeyStore alias password.
     * @param tsJks       <code>String</code> TrustStore JKS.
     * @param tsPwd       <code>String</code> TrustStore password.
     * @param acceptHost  <code>String</code> accepted hostname.
     * @param connTimeout <code>long</code> connection timeout.
     * @param readTimeout <code>long</code> read timeout.
     */
    public ApitamopkiClient(final URL wsdl,
                            final String ksJks,
                            final String ksPwd,
                            final String ksAlias,
                            final String ksAliasPwd,
                            final String tsJks,
                            final String tsPwd,
                            final String acceptHost,
                            final long connTimeout,
                            final long readTimeout) {

        if (log.isDebugEnabled()) {
            log.debug("ApitamopkiClient() - init. wsdl=" + wsdl.getFile() + ", ksJks=" + ksJks + ", ksAlias=" + ksAlias +
                      ", tsJks=" + tsJks + ", acceptedHost=" + acceptHost +
                      ", connTimeout=" + connTimeout + ", readTimeout=" + readTimeout);
        }

        // WSDL
        this.wsdl = wsdl;
        // KeyManager
        this.ksJks = ksJks;
        this.ksPwd = ksPwd;
        this.ksAlias = ksAlias;
        this.ksAliasPwd = ksAliasPwd;
        // TrustManager
        this.tsJks = tsJks;
        this.tsPwd = tsPwd;
        // Hostname to accept
        this.acceptHost = acceptHost;
        // Timeouts
        this.connTimeout = connTimeout;
        this.readTimeout = readTimeout;

        sslDebug();
    }

    /**
     * Enable / disable SSL debugging
     */
    private void sslDebug() {

        if (log.isTraceEnabled()) {
            //System.setProperty("javax.net.debug", "all:verbose");
            System.setProperty("javax.net.debug", "ssl:handshake:verbose");
            log.trace("sslDebug() - TRACE: " + System.getProperty("javax.net.debug"));
        } else if (System.getProperty("javax.net.debug") != null) {
            System.clearProperty("javax.net.debug");
        }
    }

    /**
     * Creates the service.
     * </p>
     *
     * @return ApitamopkiType object.
     */
    private ApitamopkiType getService() {

        if (log.isDebugEnabled()) {
            log.debug("getService() - begin. WSDL=" + this.wsdl.getFile());
        }

        // Service
        final ApitamopkiService service = new ApitamopkiService(this.wsdl, ApitamopkiService.SERVICE);
        final ApitamopkiType port = service.getPort(ApitamopkiService.ApitamopkiPort, ApitamopkiType.class);
        setProvider(port);
        // Client
        final HTTPClientPolicy policy = getPolicy();
        final Client client = ClientProxy.getClient(port);
        final HTTPConduit conduit = (HTTPConduit)client.getConduit();
        conduit.setClient(policy);
        // TLS-parameters
        final TLSParameters tlsParams = new TLSParameters(this.ksJks,
                                                          this.ksPwd,
                                                          this.ksAlias,
                                                          this.ksAliasPwd,
                                                          this.tsJks,
                                                          this.tsPwd,
                                                          this.acceptHost);

        final TLSClientParameters tlsClientParams = tlsParams.getTLSClientParameters();
        conduit.setTlsClientParameters(tlsClientParams);
        final Bus bus = client.getBus();

        if (log.isInfoEnabled()) {
            // Turn on message logging
            bus.getInInterceptors().add(getLoggingInInterceptor());
            bus.getOutInterceptors().add(getLoggingOutInterceptor());
            // Turn on fault message logging
            bus.getInFaultInterceptors().add(getLoggingInInterceptor());
            bus.getOutFaultInterceptors().add(getLoggingOutInterceptor());
        }

        if (log.isDebugEnabled()) {
            log.debug("getService() - end.");
        }

        return port;
    }

    private LoggingInInterceptor getLoggingInInterceptor() {

        final LoggingInInterceptor lii = new LoggingInInterceptor();
        lii.setLogBinary(false);
        lii.setLogMultipart(true);
        lii.setPrettyLogging(false);

        return lii;
    }

    private LoggingOutInterceptor getLoggingOutInterceptor() {

        final LoggingOutInterceptor loi = new LoggingOutInterceptor();
        loi.setLogBinary(false);
        loi.setLogMultipart(true);
        loi.setPrettyLogging(false);

        return loi;
    }

    /**
     * Creates the HTTPClientPolicy object.
     * </p>
     *
     * @return HTTPClientPolicy object.
     */
    private HTTPClientPolicy getPolicy() {

        final HTTPClientPolicy policy = new HTTPClientPolicy();
        policy.setConnection(ConnectionType.KEEP_ALIVE);
        policy.setAllowChunking(false);
        policy.setAutoRedirect(true);
        policy.setConnectionTimeout(this.connTimeout);
        policy.setReceiveTimeout(this.readTimeout);

        return policy;
    }

    /**
     * Creates the BindingProvider object.
     * </p>
     *
     * @param port <code>ApitamopkiType</code>
     */
    private void setProvider(final ApitamopkiType port) {

        final BindingProvider provider = (BindingProvider)port;
        // MTOM attachments
        ((SOAPBinding)provider.getBinding()).setMTOMEnabled(true);
        // Maintain HTTP session. Roles are stored in the session if needed
        //final Map<String, Object> rc = provider.getRequestContext();
        //rc.put(Message.MAINTAIN_SESSION, Boolean.TRUE);
    }

    /**
     * Destorys the client and the service.
     */
    public void destroyClient() {

        if (log.isDebugEnabled()) {
            log.debug("destroyClient() - begin.");
        }

        if (this.port != null) {
            try {
                final Client client = ClientProxy.getClient(this.port);

                if (client != null) {
                    client.destroy();
                }
            } catch (Exception e) {
                log.error(ExceptionUtils.getFullStackTrace(e));
            }

            this.port = null;
        }

        if (log.isDebugEnabled()) {
            log.debug("destroyClient() - end.");
        }
    }

    /**
     * Sends the filing and/or attachments.
     * </p>
     *
     * @param language    <code>String</code> language code for messages.
     * @param emails      <code>String</code> e-mail addersses.
     * @param background  <code>boolean</code> true = background processing.
     * @param filing      <code>String</code> reporting filing.
     * @param attachments <code>Set</code> PDF attachments.
     * @param path        <code>String</code> path to the results.
     * @return ResponseDTO object.
     */
    public ResponseDTO sendOperation(final String language,
                                     final String emails,
                                     final boolean background,
                                     final String filing,
                                     final Set<String> attachments,
                                     final String path) {

        if (log.isDebugEnabled()) {
            log.debug("sendOperation() - begin. language=" + language + ", emails=" + emails + ", background=" + background +
                      ", filing=" + filing + ", attachments=" + attachments.size() + ", path=" + path);
            log.debug("\twsdl=" + this.wsdl + ", ksJks=" + this.ksJks + ", ksAlias=" + this.ksAlias +
                      ", tsJks=" + this.tsJks + ", acceptedHost=" + this.acceptHost);
        }

        final ResponseDTO dto = new ResponseDTO();
        String msg = null;
        boolean ok = false;

        try {
            final DeliveryDataSendRequest req = getSendRequest(language, emails, background, filing, attachments);

            if (checkSendOperation(dto, req)) {
                // Request parameters are ok
                if (this.port == null) {
                    this.port = getService();
                }

                dto.sendResponse = this.port.sendOperation(req);
                setSendResponse(dto, path);
                ok = true;
            } else {
                // Errors in the request parameters
                return dto;
            }
        } catch (SOAPFaultException | Fault se) {
            msg = getSoapFault(se);
        } catch (Exception e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
            msg = ExceptionUtils.getRootCauseMessage(e);
        } finally {
            dto.resCode = getResponseCode();
        }

        if (ok) {
            if (!background) {
                dto.checkupResult = getCheckupResult(dto.result);
            }
        } else {
            dto.sendResponse = new DeliveryDataSendResponse();
            dto.sendResponse.setResult(new DeliveryDataSendResponse.Result());
            dto.sendResponse.getResult().setFiling(filing);
            dto.sendResponse.getResult().setTimestamp(getTimeStamp());

            if (msg != null) {
                dto.sendResponse.getResult().setInformation(msg);
            } else {
                if (dto.resCode < 200 || dto.resCode >= 400) {
                    // Response code of the request
                    dto.sendResponse.getResult().setInformation("Lhetyksen paluukoodi: " + dto.resCode);
                }
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("sendOperation() - end. ok=" + ok + ", error: " + msg);
        }

        return dto;
    }

    /**
     * Creates the send request.
     * </p>
     *
     * @param language    <code>String</code> language code.
     * @param emails      <code>String</code> email addresses.
     * @param background  <code>boolean</code> true = backgroung processing.
     * @param filing      <code>String</code> filing to send.
     * @param attachments <code>Set</code> attachments to send.
     * @return DeliveryDataSendRequest object.
     */
    private DeliveryDataSendRequest getSendRequest(final String language,
                                                   final String emails,
                                                   final boolean background,
                                                   final String filing,
                                                   final Set<String> attachments) {

        final ObjectFactory of = new ObjectFactory();
        final DeliveryDataSendRequest req = of.createDeliveryDataSendRequest();

        try {
            req.setLanguage(LanguageType.fromValue(language));
        } catch (IllegalArgumentException e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }

        if (!Utils.isEmpty(emails)) {
            req.setEmails(of.createDeliveryDataSendRequestEmails());
            req.getEmails().getEmail().addAll(getEmails(emails));
        }

        req.setBackgroundProcessing(background);
        req.setReportingMaterial(getFiling(filing));

        if (attachments != null && !attachments.isEmpty()) {
            req.setAttachments(new DeliveryDataSendRequest.Attachments());
            req.getAttachments().getAttachment().addAll(getAttachments(attachments));

            if (log.isDebugEnabled()) {
                log.debug("getSendRequest() - Attachments: " + attachments);
            }
        }

        return req;
    }

    /**
     * Retrieves the response filing or the checkup result.
     * </p>
     *
     * @param language    <code>String</code> language code for messages.
     * @param retrievalId <code>String</code> ID of the response filing.
     * @param resultId    <code>String</code> ID of the checkup result.
     * @param path        <code>String</code> path to the results.
     * @return ResponseDTO object.
     */
    public ResponseDTO retrievalOperation(final String language,
                                          final String retrievalId,
                                          final String resultId,
                                          final String path) {

        if (log.isDebugEnabled()) {
            log.debug("retrievalOperation() - begin. language=" + language + ", retrievalId=" + retrievalId +
                      ", resultId=" + resultId + ", path=" + path);
            log.debug("\twsdl=" + this.wsdl + ", ksJks=" + this.ksJks + ", ksAlias=" + this.ksAlias +
                      ", tsJks=" + this.tsJks + ", acceptedHost=" + this.acceptHost);
        }

        final ResponseDTO dto = new ResponseDTO();
        String msg = null;
        boolean ok = false;

        try {
            final DeliveryDataRetrievalRequest req = getRetrievalRequest(language, retrievalId, resultId);

            if (checkRetrievalOperation(dto, req)) {
                // Request parameters are ok
                if (this.port == null) {
                    this.port = getService();
                }

                dto.retrResponse = this.port.retrievalOperation(req);
                setRetrievalResponse(dto, path);
                ok = true;
            } else {
                // Errors in the request parameters
                return dto;
            }
        } catch (SOAPFaultException | Fault se) {
            msg = getSoapFault(se);
        } catch (Exception e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
            msg = ExceptionUtils.getRootCauseMessage(e);
        } finally {
            dto.resCode = getResponseCode();
        }

        if (ok) {
            if (!Utils.isEmpty(resultId)) {
                dto.checkupResult = getCheckupResult(dto.result);
            }
        } else {
            dto.retrResponse = new DeliveryDataRetrievalResponse();
            dto.retrResponse.setFiling("");
            dto.retrResponse.setTimestamp(getTimeStamp());
            dto.retrResponse.setStatus(StatusType.ERROR);
            dto.retrResponse.setRetrievalId(retrievalId);
            dto.retrResponse.setResultId(resultId);

            if (msg != null) {
                dto.retrResponse.setInformation(msg);
            } else {
                if (dto.resCode < 200 || dto.resCode >= 400) {
                    // Response code of the request
                    dto.retrResponse.setInformation("Lhetyksen paluukoodi: " + dto.resCode);
                }
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("retrievalOperation() - end. ok=" + ok + ", error: " + msg);
        }

        return dto;
    }

    /**
     * Creates the retrieval request.
     * </p>
     *
     * @param language    <code>String</code> language code.
     * @param retrievalId <code>String</code> ID of the response filing.
     * @param resultId    <code>String</code> ID of the checkup result.
     * @return DeliveryDataRetrievalRequest object.
     */
    private DeliveryDataRetrievalRequest getRetrievalRequest(final String language,
                                                             final String retrievalId,
                                                             final String resultId) {

        if (log.isDebugEnabled()) {
            log.debug("retrievalOperation() - begin. language='" + language + "', retrievalId='" + retrievalId +
                      "', resultId='" + resultId + "'");
        }

        final ObjectFactory of = new ObjectFactory();
        final DeliveryDataRetrievalRequest req = of.createDeliveryDataRetrievalRequest();

        try {
            req.setLanguage(LanguageType.fromValue(language));
        } catch (IllegalArgumentException e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }

        req.setRetrievalId(retrievalId);
        req.setResultId(resultId);

        return req;
    }

    /**
     * Returns the filing to send.
     * </p>
     *
     * @param filing <code>String</code> filing file to read.
     * @return Filing to add or null if file was not found.
     */
    private DataHandler getFiling(final String filing) {

        if (log.isDebugEnabled()) {
            log.debug("getFiling() - begin. filing=" + filing);
        }

        DataHandler dh = null;

        if (!Utils.isEmpty(filing)) {
            File file = new File(filing);

            if (file.exists()) {
                dh = new DataHandler(new FileDataSource(file));
            } else {
                log.error("getFiling() - File not found: " + file);
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("getFiling() - end. type=" + (dh != null ? dh.getContentType() : ""));
        }

        return dh;
    }

    /**
     * Returns attachments to send.
     * </p>
     *
     * @param attachments <code>Set</code> attachments to add to the request.
     * @return List of attachments to send.
     */
    private List<DataHandler> getAttachments(final Set<String> attachments) {

        if (log.isDebugEnabled()) {
            log.debug("getAttachments() - begin. attachments=" + attachments.size());
        }

        final List<DataHandler> list = new ArrayList<>(attachments.size());

        for (String attachment : attachments) {
            File file = new File(attachment);

            if (file.exists()) {
                DataHandler dh = new DataHandler(new FileDataSource(file));

                if (!list.contains(dh)) {
                    list.add(new DataHandler(new FileDataSource(file)));
                } else {
                    log.info("getAttachments() - Attachment already exists: " + file);
                }
            } else {
                log.error("getAttachments() - Attachment not found: " + file);
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("getAttachments() - end. accepted=" + list.size());
        }

        return list;
    }

    /**
     * Handles the send operation response.
     * </p>
     *
     * @param dto  <code>ResponseDTO</code>
     * @param path <code>String</code>
     */
    private void setSendResponse(final ResponseDTO dto, final String path) {

        if (log.isDebugEnabled()) {
            log.debug("setSendResponse() - begin.");
        }

        if (dto.sendResponse != null) {
            if (dto.sendResponse.getCheckupResult() != null) {
                if ((dto.result = saveResult(path, dto.sendResponse.getCheckupResult())) == null) {
                    if (Utils.isEmpty(dto.sendResponse.getResult().getInformation())) {
                        // Failed to save the file
                        String msg = "Tiedoston tallennus eponnistui: " + dto.sendResponse.getCheckupResult().getName();
                        dto.sendResponse.getResult().setInformation(msg);
                    }
                }
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("setSendResponse() - end.");
        }
    }

    /**
     * Handles the retrieval operation response.
     * </p>
     *
     * @param dto  <code>ResponseDTO</code>
     * @param path <code>String</code>
     */
    private void setRetrievalResponse(final ResponseDTO dto, final String path) {

        if (log.isDebugEnabled()) {
            log.debug("setRetrievalResponse() - begin.");
        }

        if (dto.retrResponse != null) {
            if (dto.retrResponse.getDeliveryData() != null) {
                if ((dto.result = saveResult(path, dto.retrResponse.getDeliveryData())) == null) {
                    if (Utils.isEmpty(dto.retrResponse.getInformation())) {
                        // Failed to save the file
                        String msg = "Tiedoston tallennus eponnistui: " + dto.retrResponse.getDeliveryData().getName();
                        dto.retrResponse.setInformation(msg);
                    }
                }
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("setRetrievalResponse() - end.");
        }
    }

    /**
     * Saves the file.
     * </p>
     *
     * @param path <code>String</code> path to save the file.
     * @param dh   <code>DataHandler</code> file to save.
     * @return Saved file or null if failed to save the file.
     */
    private File saveResult(final String path, final DataHandler dh) {

        if (log.isDebugEnabled()) {
            log.debug("saveResult() - begin. path=" + path);
        }

        File file = null;

        if (dh != null) {
            BufferedOutputStream bos = null;

            try {
                file = new File(path, dh.getName());
                bos = new BufferedOutputStream(new FileOutputStream(file));
                dh.writeTo(bos);
                bos.flush();
            } catch (Exception e) {
                log.error(ExceptionUtils.getFullStackTrace(e));
                file = null;
            } finally {
                if (bos != null) {
                    try {
                        bos.close();
                    } catch (IOException ignore) {
                    }
                }
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("saveResult() - end. file=" + file);
        }

        return file;
    }

    /**
     * Handles SOAP exceptions.
     * </p>
     *
     * @param e <code>Exception</code> SOAPFaultException or SoapFault object.
     * @return Error message of the exception or null if fault was not found.
     */
    private String getSoapFault(final Exception e) {

        if (log.isDebugEnabled()) {
            log.debug("getSoapFault() -alkaa. Class=" + e.getClass().getSimpleName());
        }

        String msg;

        if (e instanceof SOAPFaultException) {
            SOAPFault fault = ((SOAPFaultException)e).getFault();
            msg = fault.getFaultCode() + ": " + fault.getFaultString();
        } else if (e instanceof Fault) {
            String code = ((Fault)e).getCode();
            msg = code + ": " + e.getMessage();
        } else {
            msg = null;
            log.error("getSoapFault() - Invalid instance: " + e.getClass().getSimpleName());
        }

        if (log.isDebugEnabled()) {
            log.debug("getSoapFault(): " + msg);
        }

        return msg;
    }

    /**
     * Creates XML timestamp.
     * </p>
     *
     * @return XML timestamp.
     */
    private XMLGregorianCalendar getTimeStamp() {

        XMLGregorianCalendar cal = null;

        try {
            cal = DatatypeFactory.newInstance().newXMLGregorianCalendar(new GregorianCalendar());
        } catch (DatatypeConfigurationException e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }

        return cal;
    }

    /**
     * Returns the response code of the HTTP request.
     * </p>
     *
     * @return Response code of the HTTP request or -1 if response code was not found.
     */
    private int getResponseCode() {

        if (log.isDebugEnabled()) {
            log.debug("getResponseCode() - begin.");
        }

        int code = -1;

        if (this.port != null) {
            try {
                final Client client = ClientProxy.getClient(this.port);
                final Map<String, Object> ctx = client.getResponseContext();
                code = (int)ctx.getOrDefault(Message.RESPONSE_CODE, code);
            } catch (Exception e) {
                log.error(ExceptionUtils.getFullStackTrace(e));
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("getResponseCode() - end. code=" + code);
        }

        return code;
    }

    /**
     * Checks the send operation parameters.
     * </p>
     *
     * @param dto <code>ResponseDTO</code>
     * @return true = parameters are correct else false.
     */
    private boolean checkSendOperation(final ResponseDTO dto, final DeliveryDataSendRequest req) {

        boolean ok;
        String msg;

        // Reroting material and/or attachments
        if (!(ok = req.getReportingMaterial() != null || (req.getAttachments() != null && !req.getAttachments().getAttachment().isEmpty()))) {
            // No material or attachmens are selected
            dto.errorField = "materialBP";
            msg = "Lhetyksess ei ole ilmoitussislt tai liitteit";
        } else {
            // Language code
            // Language code
            if (!(ok = (msg = checkLanguage(req.getLanguage())) == null)) {
                dto.errorField = "sendLanguageBP";
            }
        }

        if (ok) {
            if (req.isBackgroundProcessing()) {
                // E-mail addresses
                if (!(ok = req.getEmails() != null && !req.getEmails().getEmail().isEmpty())) {
                    // Background prosessing requires at least one e-mail-address
                    msg = "Taustaksittelyn yhteydess on annettava ainakin yksi shkpostiosoite";
                } else {
                    for (String email : req.getEmails().getEmail()) {
                        if (!(ok = isEmailValid(email))) {
                            // Invalid e-mail adderss
                            msg = "Shkpostiosoite on virheellinen";
                            break;
                        }
                    }
                }

                if (!ok) {
                    dto.errorField = "emailBP";
                } else {
                    if (!(ok = req.getAttachments() == null || req.getAttachments().getAttachment().isEmpty())) {
                        // If background prosessing is selected attacments can not be selected
                        dto.errorField = "attachButtonBP";
                        msg = "Taustaksittelyn yhteydess ei voi lhett liitteit";
                        ok = false;
                    }
                }
            }
        }

        if (!ok) {
            dto.sendResponse = new DeliveryDataSendResponse();
            dto.sendResponse.setResult(new DeliveryDataSendResponse.Result());
            dto.sendResponse.getResult().setAccepted(false);
            dto.sendResponse.getResult().setInformation(msg);
            log.error("checkSendOperation() - " + dto.sendResponse.getResult().getInformation());
        }

        return ok;
    }

    /**
     * Checks the retrieval operation parameters.
     * </p>
     *
     * @param dto <code>ResponseDTO</code>
     * @return true = parameters are correct else false.
     */
    private boolean checkRetrievalOperation(final ResponseDTO dto, final DeliveryDataRetrievalRequest req) {

        boolean ok;
        String msg;

        if (!(ok = !Utils.isEmpty(req.getRetrievalId()) || !Utils.isEmpty(req.getResultId()))) {
            // Retrieval ID is empty
            dto.errorField = "retrIdBP";
            msg = "Noutotunnistetta ei ole annettu";
        } else if (!(ok = Utils.isEmpty(req.getRetrievalId()) || Utils.isEmpty(req.getResultId()))) {
            // Both cannot be given at the same time
            dto.errorField = "retrIdBP";
            msg = "Nouto- ja tulostunnistetta ei voi antaa samanaikaisesti";
        } else {
            // Language code
            if (!(ok = (msg = checkLanguage(req.getLanguage())) == null)) {
                dto.errorField = "retrLanguageBP";
            }
        }

        if (!ok) {
            dto.retrResponse = new DeliveryDataRetrievalResponse();
            dto.retrResponse.setStatus(StatusType.ERROR);
            dto.retrResponse.setInformation(msg);
            log.error("checkRetrievalOperation() - " + dto.retrResponse.getInformation());
        }

        return ok;
    }

    /**
     * Checks the language code.
     * </p>
     *
     * @param lang <code>LanguageType</code>
     * @return null = ok, else an error message.
     */
    private String checkLanguage(final LanguageType lang) {

        String msg = null;

        try {
            LanguageType.fromValue(lang.value());
        } catch (IllegalArgumentException e) {
            // Invalid language code
            msg = "Virheellinen kielikoodi";
        }

        return msg;
    }

    /**
     * Returns email addresses to add to the request.
     * </p>
     *
     * @param emails <code>String</code> email addresses to add.
     * @return Set of the email addresses.
     */
    private Set<String> getEmails(final String emails) {

        final Set<String> set = new HashSet<>();

        if (!Utils.isEmpty(emails)) {
            if (emails.indexOf(';') != -1) {
                set.addAll(Arrays.asList(emails.split(";")));
            } else if (emails.indexOf(',') != -1) {
                set.addAll(Arrays.asList(emails.split(",")));
            } else {
                set.add(emails);
            }
        }

        return set;
    }

    /**
     * Email address validation.
     * </p>
     *
     * @param email <code>Sring</code> email address.
     * @return <code>boolean</code> true if email address is valid otherwise false.
     */
    private boolean isEmailValid(final String email) {

        return !Utils.isEmpty(email) && Pattern.matches(EMAIL_PATTERN, email);
    }

    /**
     * Creates the CheckupResult object from the given file.
     * </p>
     *
     * @param resultFile <code>File</code> file containing checkup result.
     * @return CheckupResult object.
     */
    private CheckupResult getCheckupResult(final File resultFile) {

        if (log.isDebugEnabled()) {
            log.debug("getCheckupResult() - begin. resultFile=" + resultFile);
        }

        CheckupResult checkupResult = null;
        final TamoResult tamoResult = getTamoResult(resultFile);

        if (tamoResult != null) {
            if (tamoResult.getCheckupResult() != null) {
                checkupResult = tamoResult.getCheckupResult();
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("getCheckupResult() - end. ok=" + (checkupResult != null));
        }

        return checkupResult;
    }

    /**
     * Creates the TamoResult object from the given file.
     * </p>
     *
     * @param resultFile <code>File</code> file containing checkup result.
     * @return TamoResult object.
     */
    private TamoResult getTamoResult(final File resultFile) {

        if (log.isDebugEnabled()) {
            log.debug("getTamoResult() - begin. resultFile=" + resultFile);
        }

        TamoResult tamoResult = null;

        if (resultFile != null && resultFile.exists()) {
            BufferedInputStream bis = null;

            try {
                final JAXBContext jc = JAXBContext.newInstance(TamoResult.class);
                final Unmarshaller unmarshaller = jc.createUnmarshaller();
                bis = new BufferedInputStream(new FileInputStream(resultFile));
                tamoResult = (TamoResult)unmarshaller.unmarshal(bis);
            } catch (Exception e) {
                log.error(ExceptionUtils.getFullStackTrace(e));
            } finally {
                if (bis != null) {
                    try {
                        bis.close();
                    } catch (IOException ignore) {
                    }
                }
            }
        } else {
            log.error("getTamoResult() - File not found: " + resultFile);
        }

        if (log.isDebugEnabled()) {
            log.debug("getTamoResult() - begin. ok=" + (resultFile != null));
        }

        return tamoResult;
    }

    public void setWsdl(URL wsdl) {
        this.wsdl = wsdl;
    }

    public void setKsJks(String ksJks) {
        this.ksJks = ksJks;
    }

    public void setKsPwd(String ksPwd) {
        this.ksPwd = ksPwd;
    }

    public void setKsAlias(String ksAlias) {
        this.ksAlias = ksAlias;
    }

    public void setKsAliasPwd(String ksAliasPwd) {
        this.ksAliasPwd = ksAliasPwd;
    }

    public void setTsJks(String tsJks) {
        this.tsJks = tsJks;
    }

    public void setTsPwd(String tsPwd) {
        this.tsPwd = tsPwd;
    }

    public void setAcceptHost(String acceptHost) {
        this.acceptHost = acceptHost;
    }

    public void setConnTimeout(long connTimeout) {
        this.connTimeout = connTimeout;
    }

    public void setReadTimeout(long readTimeout) {
        this.readTimeout = readTimeout;
    }
}
