package org.verohallinto.apitamoclient.ui;

import ch.qos.logback.classic.Level;
import org.apache.pivot.wtk.Menu;
import org.apache.pivot.wtk.MenuItemSelectionListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.verohallinto.apitamoclient.ApitamopkiClient;
import org.verohallinto.apitamoclient.dto.CertDTO;
import org.verohallinto.apitamoclient.util.Const;
import org.verohallinto.apitamoclient.util.Props;
import org.verohallinto.apitamoclient.util.Utils;

import java.security.KeyStore;
import java.security.cert.X509Certificate;
import java.text.SimpleDateFormat;

/**
 * Main menu.
 */
public class MainMenu {

    private static final Logger log = LoggerFactory.getLogger(MainMenu.class);
    private ApitamopkiClientUI ui;
    private SimpleDateFormat sdf;

    /**
     * Constuctor.
     * </p>
     *
     * @param ui <code>ApitamopkiClientUI</code>
     */
    public MainMenu(final ApitamopkiClientUI ui) {

        this.ui = ui;
        this.sdf = new SimpleDateFormat("dd.MM.yyyy");

        // Ympristn valinta
        ui.menuEnv.getMenuItemSelectionListeners().add(new MenuItemSelectionListener() {
            @Override
            public void itemSelected(final Menu.Item item) {

                int uusiYmparisto;
                final String text = (String)item.getButtonData();

                switch (text) {
                    case Const.TITLE_DEV:
                        uusiYmparisto = Const.DEVELOPMENT;
                        break;
                    case Const.TITLE_DEVTEST:
                        uusiYmparisto = Const.DEVELOPMENT_TEST;
                        break;
                    case Const.TITLE_TEST:
                        uusiYmparisto = Const.PRODUCTION_TEST;
                        break;
                    case Const.TITLE_PROD:
                        uusiYmparisto = Const.PRODUCTION;
                        break;
                    default:
                        uusiYmparisto = ui.env;
                        break;
                }

                if (uusiYmparisto != ui.env) {
                    ui.env = uusiYmparisto;
                    setEnvironment();
                }
            }
        });

        // Varmenteen valinta
        ui.menuCrt.getMenuItemSelectionListeners().add(new MenuItemSelectionListener() {
            @Override
            public void itemSelected(final Menu.Item item) {

                ui.ksAlias = (String)item.getButtonData();
                ui.crtInfo.setButtonData(getCertInfo(ui.ksAlias));

                if (ui.apitamo != null) {
                    ui.apitamo.destroyClient();
                }

                ui.apitamo = new ApitamopkiClient(ui.wsdl,
                                                  ui.ksJks,
                                                  ui.ksPwd,
                                                  ui.ksAlias,
                                                  ui.ksAliasPwd,
                                                  ui.tsJks,
                                                  ui.tsPwd,
                                                  ui.acceptHost,
                                                  ui.connTimeout,
                                                  ui.readTimeout);
            }
        });

        // Lokitustason valinta. Oletuksesksi INFO
        final Menu.Section section = ui.menuLog.getSections().get(0);
        final Menu.Item item = section.get(2);
        item.setActive(true);
        setLogLevel(Level.toLevel((String)item.getButtonData()));

        ui.menuLog.getMenuItemSelectionListeners().add(new MenuItemSelectionListener() {
            @Override
            public void itemSelected(final Menu.Item item) {
                setLogLevel(Level.toLevel((String)item.getButtonData()));

                if (ui.apitamo != null) {
                    ui.apitamo.destroyClient();
                }

                ui.apitamo = new ApitamopkiClient(ui.wsdl,
                                                  ui.ksJks,
                                                  ui.ksPwd,
                                                  ui.ksAlias,
                                                  ui.ksAliasPwd,
                                                  ui.tsJks,
                                                  ui.tsPwd,
                                                  ui.acceptHost,
                                                  ui.connTimeout,
                                                  ui.readTimeout);
            }
        });

        setEnvironment();
        ui.getWindow().setTitle(Const.TITLES[ui.env]);
    }

    /**
     * Ympristn mukaisen varmenteen tietojen asetus.
     */
    private void setEnvironment() {

        Props.init(Const.ENV.get(ui.env));
        ui.getWindow().setTitle(Const.TITLES[ui.env]);

        try {
            ui.wsdl = Utils.getWsdl();
        } catch (Exception e) {
            ui.action.handleException(e);
        }

        ui.ksJks = Props.get("key.store.jks");
        ui.ksPwd = Props.get("key.store.pwd");
        ui.ksAliasPwd = Props.get("key.store.alias.pwd");
        ui.tsJks = Props.get("trust.store.jks");
        ui.tsPwd = Props.get("trust.store.pwd");
        ui.acceptHost = Props.get("accept.host");
        ui.connTimeout = Props.getTimeoutConn();
        ui.readTimeout = Props.getTimeoutRead();

        final Menu.Section section = ui.menuCrt.getSections().get(0);

        if (section.getLength() > 0) {
            section.remove(0, section.getLength());
        }

        final String aliases = Props.get("key.store.aliases");

        if (!Utils.isEmpty(aliases)) {
            for (String alias : aliases.split("\\|")) {
                section.add(new Menu.Item(alias));
            }
        } else {
            final String alias = Props.get("key.store.alias");
            section.add(new Menu.Item(alias));
        }

        ui.ksAlias = (String)section.get(0).getButtonData();
        ui.crtInfo.setButtonData(getCertInfo(ui.ksAlias));

        if (ui.apitamo != null) {
            ui.apitamo.destroyClient();
        }

        ui.apitamo = new ApitamopkiClient(ui.wsdl,
                                          ui.ksJks,
                                          ui.ksPwd,
                                          ui.ksAlias,
                                          ui.ksAliasPwd,
                                          ui.tsJks,
                                          ui.tsPwd,
                                          ui.acceptHost,
                                          ui.connTimeout,
                                          ui.readTimeout);
    }

    /**
     * Varmenteen tietojen palautus.
     * </p>
     *
     * @param alias <code>String</code> varmenteen alias.
     * @return Varmenteen tiedot.
     */
    private String getCertInfo(final String alias) {

        final StringBuilder sb = new StringBuilder(128);
        final CertDTO dto = getCertificate(alias);

        if (dto.getCommonName() != null && !dto.getCommonName().isEmpty()) {
            sb.append(dto.getCommonName());
        }

        if (dto.getOrganization() != null && !dto.getOrganization().isEmpty()) {
            if (sb.length() > 0) sb.append(", ");
            sb.append(dto.getOrganization());
        }

        if (dto.getValidFrom() != null && !dto.getValidFrom().isEmpty() &&
            dto.getValidTo() != null && !dto.getValidTo().isEmpty()) {
            sb.append(" (").append(dto.getValidFrom()).append("-").append(dto.getValidTo()).append(')');
        }

        return sb.length() > 0 ? sb.toString() : "Varmennetta ei lydy";
    }

    /**
     * Varmenteen luku.
     * </p>
     *
     * @param alias <code>String</code> varmenteen alias.
     * @return Varmenteen tiedot.
     */
    private CertDTO getCertificate(String alias) {

        if (log.isDebugEnabled()) {
            log.debug("getCertificate() - begin. alias=" + alias);
        }

        CertDTO dto = new CertDTO();
        final KeyStore ks = Utils.getKeyStore(ui.ksJks, ui.ksPwd);

        if (ks != null) {
            if (Utils.isEmpty(alias)) {
                alias = Props.get("key.store.alias");
            }

            try {
                final X509Certificate crt = (X509Certificate)ks.getCertificate(alias);

                if (crt != null) {
                    dto = getAttributes(crt.getSubjectDN().getName());
                    dto.setValidFrom(this.sdf.format(crt.getNotBefore()));
                    dto.setValidTo(this.sdf.format(crt.getNotAfter()));
                } else {
                    dto = new CertDTO();
                }
            } catch (Exception e) {
                ui.action.handleException(e);
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("getCertificate() - end: " + dto.toString());
        }

        return dto;
    }

    /**
     * Subject-attribuutit.
     * </p>
     * CN: CommonName
     * OU: OrganizationalUnit
     * O: Organization
     * L: Locality
     * S: StateOrProvinceName
     * C: CountryName
     *
     * @param s <code>String</code> attribuutit (CN=Sample Cert, OU=R&D, O=Company Ltd., L=Dublin 4, S=Dublin, C=IE).
     * @return Subject-kentn attribuutit.
     */
    private CertDTO getAttributes(final String s) {

        if (log.isDebugEnabled()) {
            log.debug("getAttributes() - begin. s=" + s);
        }

        final CertDTO dto = new CertDTO();

        if (s != null && !s.isEmpty()) {
            dto.setCommonName(getAttribute("CN", s));
            dto.setOrganizationalUnit(getAttribute("OU", s));
            dto.setOrganization(getAttribute("O", s));
            dto.setLocality(getAttribute("L", s));
            dto.setState(getAttribute("S", s));
            dto.setCountryName(getAttribute("C", s));
            dto.setSerialnumber(getAttribute("SERIALNUMBER", s));
        }

        if (log.isDebugEnabled()) {
            log.debug("getAttributes() - end.");
        }

        return dto;
    }

    /**
     * Attribuutin arvon haku.
     * </p>
     *
     * @param name   <code>String</code> attribuutin nimi.
     * @param values <code>String</code> attribuutit.
     * @return attribuutin arvo jos lytyy, muuten null.
     */
    private String getAttribute(final String name, final String values) {

        String value = null;
        final String haku = name.concat("=");
        int x, y;

        if ((x = values.indexOf(haku)) != -1) {
            x += haku.length();

            if ((y = values.indexOf(',', x)) == -1) {
                y = values.length();
            }

            if (y > x) {
                value = values.substring(x, y).trim();
            }
        }

        return value;
    }

    private static final String[] LOGGERS = {"org.verohallinto",
                                             "org.apache.cxf",
                                             "org.apache.pivot"};

    /**
     * Lokitustason asetus.
     * </p>
     *
     * @param level <code>Level</code> asetettava lokitustaso.
     */
    private void setLogLevel(final Level level) {

        ch.qos.logback.classic.Logger logger;

        for (String s : LOGGERS) {
            logger = (ch.qos.logback.classic.Logger)LoggerFactory.getLogger(s);
            if (logger != null) logger.setLevel(level);
        }

        ui.status.setText("Lotitustaso: " + level.toString());

        log.info("setLogLevel() - Level=" + level);
    }
}
